from __future__ import annotations  # isort: skip
import os
import shutil
import subprocess
from typing import Dict, Union

from ..constants import (CLARIFAI_EXAMPLES_REPO, CLARIFAI_EXAMPLES_REPO_PATH,
                         MODEL_UPLOAD_EXAMPLE_FOLDER)


def download_examples_repo(forced_download: bool = False):

  def _pull():
    subprocess.run(f"git clone {CLARIFAI_EXAMPLES_REPO} {CLARIFAI_EXAMPLES_REPO_PATH}")

  if not os.path.isdir(CLARIFAI_EXAMPLES_REPO_PATH):
    print(f"Download examples to {CLARIFAI_EXAMPLES_REPO_PATH}")
    _pull()
  else:
    if forced_download:

      def _rm_dir_readonly(func, path, _):
        import stat
        os.chmod(path, stat.S_IWRITE)
        func(path)

      print("Removing old examples...")
      shutil.rmtree(CLARIFAI_EXAMPLES_REPO_PATH, onerror=_rm_dir_readonly)
      _pull()


def list_model_upload_examples(
    forced_download: bool = False) -> Dict[str, tuple[str, Union[str, None]]]:
  download_examples_repo(forced_download)
  model_upload_folder = MODEL_UPLOAD_EXAMPLE_FOLDER
  model_upload_path = os.path.join(CLARIFAI_EXAMPLES_REPO_PATH, model_upload_folder)
  examples = {}
  for model_type_ex in os.listdir(model_upload_path):
    _folder = os.path.join(model_upload_path, model_type_ex)
    if os.path.isdir(_folder):
      _walk = list(os.walk(_folder))
      if len(_walk) > 0:
        _, model_names, _files = _walk[0]
        readme = [item for item in _files if "readme" in item.lower()]
        for name in model_names:
          examples.update({
              f"{model_type_ex}/{name}": [
                  os.path.join(_folder, name),
                  os.path.join(_folder, readme[0]) or None
              ]
          })

  return examples



TRAINABLE_MODEL_TYPES = [
    'visual-classifier', 'visual-detector', 'visual-segmenter', 'visual-embedder', 'clusterer',
    'text-classifier', 'embedding-classifier', 'text-to-text'
]
MAX_MODEL_PREDICT_INPUTS = 128



import logging
import os
from datetime import datetime

import pytest

from clarifai.client.user import User
from clarifai.datasets.upload.utils import load_module_dataloader

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
NOW = str(int(datetime.now().timestamp()))
CREATE_APP_ID = f"ci_input_app_{NOW}"
CREATE_DATASET_ID = "ci_input_test_dataset"

#assets
IMAGE_URL = "https://samples.clarifai.com/metro-north.jpg"
VIDEO_URL = "https://samples.clarifai.com/beer.mp4"
TEXT_URL = "https://samples.clarifai.com/featured-models/Llama2_Conversational-agent.txt"
AUDIO_URL = "https://samples.clarifai.com/english_audio_sample.mp3"
IMAGE_FILE_PATH = os.path.dirname(__file__) + "/assets/red-truck.png"
VIDEO_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.mp4"
TEXT_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.txt"
AUDIO_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.mp3"
CSV_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.csv"
FOLDER_PATH = os.path.dirname(__file__) + "/assets/test"
MODULE_DIR = os.path.dirname(__file__) + "/assets/voc"


def create_app():
  client = User(user_id=CREATE_APP_USER_ID)
  return client.create_app(app_id=CREATE_APP_ID, base_workflow="Empty")


@pytest.mark.requires_secrets
class Testdataupload:
  """Tests for data uploads.
  Uploads are tested for each of the following resources:
  - image
  - video
  - text
  - audio

  Tests for the following upload methods:
  - url
  - filepath
  - rawtext
  - CSV
  - Folder
  """

  @classmethod
  def setup_class(self):
    self.app = create_app()
    self.input_object = self.app.inputs()
    self.dataset = self.app.create_dataset(dataset_id=CREATE_DATASET_ID)

  def test_upload_image_url(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_url(input_id='input_1', image_url=IMAGE_URL)
      assert "SUCCESS" in caplog.text

  def test_upload_video_url(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_url(input_id='input_2', video_url=VIDEO_URL)
      assert "SUCCESS" in caplog.text

  def test_upload_text_url(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_url(input_id='input_3', text_url=TEXT_URL)
      assert "SUCCESS" in caplog.text

  def test_upload_audio_url(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_url(input_id='input_4', audio_url=AUDIO_URL)
      assert "SUCCESS" in caplog.text

  def test_upload_image_filepath(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_file(input_id='input_5', image_file=IMAGE_FILE_PATH)
      assert "SUCCESS" in caplog.text

  def test_upload_video_filepath(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_file(input_id='input_6', video_file=VIDEO_FILE_PATH)
      assert "SUCCESS" in caplog.text

  def test_upload_audio_filepath(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_file(input_id='input_7', audio_file=AUDIO_FILE_PATH)
      assert "SUCCESS" in caplog.text

  def test_upload_text_filepath(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_from_file(input_id='input_8', text_file=TEXT_FILE_PATH)
      assert "SUCCESS" in caplog.text

  def test_upload_rawtext(self, caplog):
    with caplog.at_level(logging.INFO):
      self.input_object.upload_text(input_id='input_9', raw_text='This is a test text')
      assert "SUCCESS" in caplog.text

  def test_list_inputs(self):
    paginated_inputs = list(self.input_object.list_inputs(page_no=1, per_page=5))
    image_filterd_inputs = list(self.input_object.list_inputs(input_type='image'))
    downloaded_inputs = self.input_object.download_inputs(image_filterd_inputs)
    assert len(downloaded_inputs) == 2  #download inputs check
    assert len(paginated_inputs) == 5
    assert len(image_filterd_inputs) == 2  # 2 images uploaded in the above tests

  def test_aggregate_inputs(self, caplog):
    uploaded_inputs = list(self.input_object.list_inputs())
    with caplog.at_level(logging.INFO):
      self.input_object.delete_inputs(uploaded_inputs)
      assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
    assert len(uploaded_inputs) == 9  # 9 inputs uploaded in the above tests

  def test_upload_csv(self, caplog):
    self.dataset.upload_from_csv(
        csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True)
    uploaded_inputs = list(self.input_object.list_inputs(dataset_id=CREATE_DATASET_ID))
    concepts = list(self.app.list_concepts())
    with caplog.at_level(logging.INFO):
      self.input_object.delete_inputs(uploaded_inputs)
      assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
    assert len(uploaded_inputs) == 5  # 5 inputs are uploaded from the CSV file
    assert len(concepts) == 2  # Test for list concepts

  def test_upload_folder(self, caplog):
    self.dataset.upload_from_folder(folder_path=FOLDER_PATH, input_type='image', labels=True)
    uploaded_inputs = list(self.input_object.list_inputs())
    with caplog.at_level(logging.INFO):
      self.input_object.delete_inputs(uploaded_inputs)
      assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
    assert uploaded_inputs[0].data.concepts[0].name == 'test'  # label of the first input in the folder
    assert len(uploaded_inputs) == 3  # 3 inputs are uploaded from the folder

  def test_upload_dataset(self, caplog):
    dataloader = load_module_dataloader(module_dir=MODULE_DIR, split="train")
    self.dataset.upload_dataset(dataloader)
    uploaded_inputs = list(self.input_object.list_inputs())
    annotations = list(self.input_object.list_annotations(batch_input=uploaded_inputs))
    with caplog.at_level(logging.INFO):
      self.input_object.delete_inputs(uploaded_inputs)
      assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
    assert len(uploaded_inputs) == 10  # 3 inputs are uploaded from the folder
    assert len(annotations) == 28  # Test for list annotatoins

  @classmethod
  def teardown_class(self):
    self.app.delete_dataset(dataset_id=CREATE_DATASET_ID)
    User(user_id=CREATE_APP_USER_ID).delete_app(app_id=CREATE_APP_ID)



import os
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from multiprocessing import cpu_count
from typing import Dict, Generator, List, Optional, Tuple, Type, TypeVar, Union

import requests
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.service_pb2 import MultiInputResponse
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf.json_format import MessageToDict
from requests.adapters import HTTPAdapter, Retry
from tqdm import tqdm

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.dataset import DATASET_UPLOAD_TASKS
from clarifai.datasets.export.inputs_annotations import (DatasetExportReader,
                                                         InputAnnotationDownloader)
from clarifai.datasets.upload.base import ClarifaiDataLoader
from clarifai.datasets.upload.image import (VisualClassificationDataset, VisualDetectionDataset,
                                            VisualSegmentationDataset)
from clarifai.datasets.upload.text import TextClassificationDataset
from clarifai.datasets.upload.utils import DisplayUploadStatus
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import add_file_handler, get_logger
from clarifai.utils.misc import BackoffIterator, Chunker

ClarifaiDatasetType = TypeVar('ClarifaiDatasetType', VisualClassificationDataset,
                              VisualDetectionDataset, VisualSegmentationDataset,
                              TextClassificationDataset)


class Dataset(Lister, BaseClient):
  """Dataset is a class that provides access to Clarifai API endpoints related to Dataset information."""

  def __init__(self,
               url: str = None,
               dataset_id: str = None,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               **kwargs):
    """Initializes a Dataset object.

    Args:
        url (str): The URL to initialize the dataset object.
        dataset_id (str): The Dataset ID within the App to interact with.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        **kwargs: Additional keyword arguments to be passed to the Dataset.
    """
    if url and dataset_id:
      raise UserError("You can only specify one of url or dataset_id.")
    if url:
      user_id, app_id, _, dataset_id, _ = ClarifaiUrlHelper.split_clarifai_url(url)
      kwargs = {'user_id': user_id, 'app_id': app_id}
    self.kwargs = {**kwargs, 'id': dataset_id}
    self.dataset_info = resources_pb2.Dataset(**self.kwargs)
    # Related to Dataset Upload
    self.num_workers: int = min(10, cpu_count())  #15 req/sec rate limit
    self.annot_num_workers = 4
    self.max_retires = 10
    self.batch_size = 128  # limit max protos in a req
    self.task = None  # Upload dataset type
    self.input_object = Inputs(user_id=self.user_id, app_id=self.app_id, pat=pat, token=token)
    self.logger = get_logger(logger_level="INFO", name=__name__)
    BaseClient.__init__(
        self, user_id=self.user_id, app_id=self.app_id, base=base_url, pat=pat, token=token)
    Lister.__init__(self)

  def create_version(self, **kwargs) -> 'Dataset':
    """Creates a dataset version for the Dataset.

    Args:
        **kwargs: Additional keyword arguments to be passed to Dataset Version.
          - description (str): The description of the dataset version.
          - metadata (dict): The metadata of the dataset version.

    Returns:
        Dataset: A Dataset object for the specified dataset ID.

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
        >>> dataset_version = dataset.create_version(description='dataset_version_description')
    """
    request = service_pb2.PostDatasetVersionsRequest(
        user_app_id=self.user_app_id,
        dataset_id=self.id,
        dataset_versions=[resources_pb2.DatasetVersion(**kwargs)])

    response = self._grpc_request(self.STUB.PostDatasetVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nDataset Version created\n%s", response.status)
    kwargs.update({
        'dataset_id': self.id,
        'version': response.dataset_versions[0],
    })

    return Dataset.from_auth_helper(self.auth_helper, **kwargs)

  def delete_version(self, version_id: str) -> None:
    """Deletes a dataset version for the Dataset.

    Args:
        version_id (str): The version ID to delete.

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
        >>> dataset.delete_version(version_id='version_id')
    """
    request = service_pb2.DeleteDatasetVersionsRequest(
        user_app_id=self.user_app_id, dataset_id=self.id, dataset_version_ids=[version_id])

    response = self._grpc_request(self.STUB.DeleteDatasetVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nDataset Version Deleted\n%s", response.status)

  def list_versions(self, page_no: int = None,
                    per_page: int = None) -> Generator['Dataset', None, None]:
    """Lists all the versions for the dataset.

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Dataset: Dataset objects for the versions of the dataset.

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
        >>> all_dataset_versions = list(dataset.list_versions())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        dataset_id=self.id,
    )
    all_dataset_versions_info = self.list_pages_generator(
        self.STUB.ListDatasetVersions,
        service_pb2.ListDatasetVersionsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for dataset_version_info in all_dataset_versions_info:
      dataset_version_info['id'] = dataset_version_info['dataset_version_id']
      del dataset_version_info['dataset_version_id']
      del dataset_version_info['metrics']
      kwargs = {
          'dataset_id': self.id,
          'version': resources_pb2.DatasetVersion(**dataset_version_info),
      }
      yield Dataset.from_auth_helper(self.auth_helper, **kwargs)

  def _concurrent_annot_upload(self, annots: List[List[resources_pb2.Annotation]]
                              ) -> Union[List[resources_pb2.Annotation], List[None]]:
    """Uploads annotations concurrently.

    Args:
      annots: annot protos

    Returns:
      retry_annot_upload: All failed annot protos during upload
    """
    annot_threads = []
    retry_annot_upload = []

    with ThreadPoolExecutor(max_workers=self.annot_num_workers) as executor:  # limit annot workers
      annot_threads = [
          executor.submit(self.input_object.upload_annotations, inp_batch, False)
          for inp_batch in annots
      ]

      for job in as_completed(annot_threads):
        result = job.result()
        if result:
          retry_annot_upload.extend(result)

    return retry_annot_upload

  def _delete_failed_inputs(
      self,
      batch_input_ids: List[int],
      dataset_obj: ClarifaiDatasetType,
      upload_response: MultiInputResponse = None) -> Tuple[List[int], List[int]]:
    """Delete failed input ids from clarifai platform dataset.

    Args:
      batch_input_ids: batch input ids
      dataset_obj: ClarifaiDataset object
      upload_response: upload response proto

    Returns:
      success_inputs: upload success input ids
      failed_inputs: upload failed input ids
    """
    success_status = status_pb2.Status(code=status_code_pb2.INPUT_DOWNLOAD_SUCCESS)
    input_ids = {dataset_obj.all_input_ids[id]: id for id in batch_input_ids}
    response = self._grpc_request(
        self.STUB.ListInputs,
        service_pb2.ListInputsRequest(
            ids=list(input_ids.keys()),
            per_page=len(input_ids),
            user_app_id=self.user_app_id,
            status=success_status),
    )
    response_dict = MessageToDict(response)
    success_inputs = response_dict.get('inputs', [])

    success_input_ids = [input.get('id') for input in success_inputs]
    failed_input_ids = list(set(input_ids) - set(success_input_ids.copy()))
    #check duplicate input ids
    duplicate_input_ids = [
        input.id for input in upload_response.inputs
        if input.status.details == 'Input has a duplicate ID.'
    ]  #handling duplicte ID failures.
    if duplicate_input_ids:
      success_input_ids = list(set(success_input_ids.copy()) - set(duplicate_input_ids.copy()))
      failed_input_ids = list(set(failed_input_ids) - set(duplicate_input_ids))
      self.logger.warning(
          f"Upload Failed for {len(duplicate_input_ids)} inputs in current batch: Duplicate input ids: {duplicate_input_ids}"
      )

    #delete failed inputs
    self._grpc_request(
        self.STUB.DeleteInputs,
        service_pb2.DeleteInputsRequest(user_app_id=self.user_app_id, ids=failed_input_ids),
    )
    return [input_ids[id] for id in success_input_ids], [input_ids[id] for id in failed_input_ids]

  def _upload_inputs_annotations(
      self, batch_input_ids: List[int], dataset_obj: ClarifaiDatasetType
  ) -> Tuple[List[int], List[resources_pb2.Annotation], MultiInputResponse]:
    """Uploads batch of inputs and annotations concurrently to clarifai platform dataset.

    Args:
      batch_input_ids: batch input ids
      dataset_obj: ClarifaiDataset object

    Returns:
      failed_input_ids: failed input ids
      retry_annot_protos: failed annot protos
      response: upload response proto
    """
    input_protos, _ = dataset_obj.get_protos(batch_input_ids)
    input_job_id, _response = self.input_object.upload_inputs(inputs=input_protos, show_log=False)
    retry_annot_protos = []

    self.input_object._wait_for_inputs(input_job_id)
    success_input_ids, failed_input_ids = self._delete_failed_inputs(batch_input_ids, dataset_obj,
                                                                     _response)

    if self.task in ["visual_detection", "visual_segmentation"] and success_input_ids:
      _, annotation_protos = dataset_obj.get_protos(success_input_ids)
      chunked_annotation_protos = Chunker(annotation_protos, self.batch_size).chunk()
      retry_annot_protos.extend(self._concurrent_annot_upload(chunked_annotation_protos))

    return failed_input_ids, retry_annot_protos, _response

  def _retry_uploads(self, failed_input_ids: List[int],
                     retry_annot_protos: List[resources_pb2.Annotation],
                     dataset_obj: ClarifaiDatasetType) -> None:
    """Retry failed uploads.

    Args:
      failed_input_ids: failed input ids
      retry_annot_protos: failed annot protos
      dataset_obj: ClarifaiDataset object
    """
    if failed_input_ids:
      retry_input_ids = [dataset_obj.all_input_ids[id] for id in failed_input_ids]
      #Log Retrying inputs
      self.logger.warning(
          f"Retrying upload for {len(failed_input_ids)} inputs in current batch: {retry_input_ids}"
      )
      failed_retrying_inputs, _, retry_response = self._upload_inputs_annotations(
          failed_input_ids, dataset_obj)
      #Log failed inputs
      if failed_retrying_inputs:
        failed_retrying_input_ids = [
            dataset_obj.all_input_ids[id] for id in failed_retrying_inputs
        ]
        failed_inputs_logs = {
            input.id: input.status.details
            for input in retry_response.inputs if input.id in failed_retrying_input_ids
        }
        self.logger.warning(
            f"Failed to upload {len(failed_retrying_inputs)} inputs in current batch: {failed_inputs_logs}"
        )
    if retry_annot_protos:
      chunked_annotation_protos = Chunker(retry_annot_protos, self.batch_size).chunk()
      _ = self._concurrent_annot_upload(chunked_annotation_protos)

  def _data_upload(self, dataset_obj: ClarifaiDatasetType) -> None:
    """Uploads inputs and annotations to clarifai platform dataset.

    Args:
      dataset_obj: ClarifaiDataset object
    """
    input_ids = list(range(len(dataset_obj)))
    chunk_input_ids = Chunker(input_ids, self.batch_size).chunk()
    with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
      with tqdm(total=len(chunk_input_ids), desc='Uploading Dataset') as progress:
        # Submit all jobs to the executor and store the returned futures
        futures = [
            executor.submit(self._upload_inputs_annotations, batch_input_ids, dataset_obj)
            for batch_input_ids in chunk_input_ids
        ]

        for job in as_completed(futures):
          retry_input_ids, retry_annot_protos, _ = job.result()
          self._retry_uploads(retry_input_ids, retry_annot_protos, dataset_obj)
          progress.update()

  def upload_dataset(self,
                     dataloader: Type[ClarifaiDataLoader],
                     batch_size: int = 32,
                     get_upload_status: bool = False,
                     log_warnings: bool = False) -> None:
    """Uploads a dataset to the app.

    Args:
      dataloader (Type[ClarifaiDataLoader]): ClarifaiDataLoader object
      batch_size (int): batch size for concurrent upload of inputs and annotations (max: 128)
      get_upload_status (bool): True if you want to get the upload status of the dataset
      log_warnings (bool): True if you want to save log warnings in a file
    """
    #add file handler to log warnings
    if log_warnings:
      add_file_handler(self.logger, f"Dataset_Upload{str(int(datetime.now().timestamp()))}.log")
    #set batch size and task
    self.batch_size = min(self.batch_size, batch_size)
    self.task = dataloader.task
    if self.task not in DATASET_UPLOAD_TASKS:
      raise UserError("Task should be one of \
                      'text_classification', 'visual_classification', \
                      'visual_detection', 'visual_segmentation', 'visual_captioning'")

    if self.task == "text_classification":
      dataset_obj = TextClassificationDataset(dataloader, self.id)

    elif self.task == "visual_detection":
      dataset_obj = VisualDetectionDataset(dataloader, self.id)

    elif self.task == "visual_segmentation":
      dataset_obj = VisualSegmentationDataset(dataloader, self.id)

    else:  # visual_classification & visual_captioning
      dataset_obj = VisualClassificationDataset(dataloader, self.id)

    if get_upload_status:
      pre_upload_stats = self.get_upload_status(pre_upload=True)

    self._data_upload(dataset_obj)

    if get_upload_status:
      self.get_upload_status(dataloader=dataloader, pre_upload_stats=pre_upload_stats)

  def upload_from_csv(self,
                      csv_path: str,
                      input_type: str = 'text',
                      csv_type: str = None,
                      labels: bool = True,
                      batch_size: int = 128) -> None:
    """Uploads dataset from a csv file.

    Args:
        csv_path (str): path to the csv file
        input_type (str): type of the dataset(text, image)
        csv_type (str): type of the csv file(raw, url, file_path)
        labels (bool): True if csv file has labels column
        batch_size (int): batch size for concurrent upload of inputs and annotations

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> dataset = Dataset(user_id = 'user_id', app_id = 'demo_app', dataset_id = 'demo_dataset')
        >>> dataset.upload_from_csv(csv_path='csv_path', input_type='text', csv_type='raw, labels=True)

    Note:
        CSV file supports 'inputid', 'input', 'concepts', 'metadata', 'geopoints' columns.
        All the data in the CSV should be in double quotes.
        metadata should be in single quotes format. Example: "{'key': 'value'}"
        geopoints should be in "long,lat" format.
    """
    if input_type not in ['image', 'text', 'video', 'audio']:
      raise UserError('Invalid input type, it should be image,text,audio or video')
    if csv_type not in ['raw', 'url', 'file_path']:
      raise UserError('Invalid csv type, it should be raw, url or file_path')
    assert csv_path.endswith('.csv'), 'csv_path should be a csv file'
    if csv_type == 'raw' and input_type != 'text':
      raise UserError('Only text input type is supported for raw csv type')
    batch_size = min(128, batch_size)
    input_protos = self.input_object.get_inputs_from_csv(
        csv_path=csv_path,
        input_type=input_type,
        csv_type=csv_type,
        dataset_id=self.id,
        labels=labels)
    self.input_object._bulk_upload(inputs=input_protos, batch_size=batch_size)

  def upload_from_folder(self,
                         folder_path: str,
                         input_type: str,
                         labels: bool = False,
                         batch_size: int = 128) -> None:
    """Upload dataset from folder.

    Args:
        folder_path (str): Path to the folder containing images.
        input_type (str): type of the dataset(text, image)
        labels (bool): True if folder name is the label for the inputs
        batch_size (int): batch size for concurrent upload of inputs and annotations

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> dataset = Dataset(user_id = 'user_id', app_id = 'demo_app', dataset_id = 'demo_dataset')
        >>> dataset.upload_from_folder(folder_path='folder_path', input_type='text', labels=True)

    Note: The filename is used as the input_id.
    """
    if input_type not in ['image', 'text']:
      raise UserError('Invalid input type it should be image or text')
    if input_type == 'image':
      input_protos = self.input_object.get_image_inputs_from_folder(
          folder_path=folder_path, dataset_id=self.id, labels=labels)
    if input_type == 'text':
      input_protos = self.input_object.get_text_inputs_from_folder(
          folder_path=folder_path, dataset_id=self.id, labels=labels)
    self.input_object._bulk_upload(inputs=input_protos, batch_size=batch_size)

  def get_upload_status(
      self,
      dataloader: Type[ClarifaiDataLoader] = None,
      delete_version: bool = False,
      timeout: int = 600,
      pre_upload_stats: Tuple[Dict[str, int], Dict[str, int]] = None,
      pre_upload: bool = False) -> Optional[Tuple[Dict[str, int], Dict[str, int]]]:
    """Creates a new dataset version and displays the upload status of the dataset.

    Args:
        dataloader (Type[ClarifaiDataLoader]): ClarifaiDataLoader object
        delete_version (bool): True if you want to delete the version after getting the upload status
        timeout (int): Timeout in seconds for getting the upload status. Default is 600 seconds.
        pre_upload_stats (Tuple[Dict[str, int], Dict[str, int]]): The pre upload stats for the dataset.
        pre_upload (bool): True if you want to get the pre upload stats for the dataset.

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
        >>> dataset.get_upload_status(dataloader)

    Note:
        This is a beta feature and is subject to change.
    """
    self.logger.info("Getting dataset upload status...")
    dataset_version_id = uuid.uuid4().hex
    _ = self.create_version(id=dataset_version_id, description="SDK Upload Status")

    request_data = dict(
        user_app_id=self.user_app_id,
        dataset_id=self.id,
        dataset_version_id=dataset_version_id,
    )

    start_time = time.time()
    backoff_iterator = BackoffIterator()
    while (True):
      dataset_metrics_response = self._grpc_request(
          self.STUB.ListDatasetVersionMetricsGroups,
          service_pb2.ListDatasetVersionMetricsGroupsRequest(**request_data),
      )

      if dataset_metrics_response.status.code != status_code_pb2.SUCCESS:
        self.delete_version(dataset_version_id)
        raise Exception("Failed to get dataset metrics {}".format(dataset_metrics_response.status))

      dict_response = MessageToDict(dataset_metrics_response)
      if len(dict_response.keys()) == 1 and time.time() - start_time < timeout:
        self.logger.info("Crunching the dataset metrics. Please wait...")
        time.sleep(next(backoff_iterator))
        continue
      else:
        if time.time() - start_time > timeout:
          self.delete_version(dataset_version_id)
          raise UserError(
              "Dataset metrics are taking too long to process. Please try again later.")
        break
    #get pre upload stats
    if pre_upload:
      return DisplayUploadStatus.get_dataset_version_stats(dataset_metrics_response)

    dataset_info_dict = dict(user_id=self.user_id, app_id=self.app_id, dataset_id=self.id)
    DisplayUploadStatus(dataloader, dataset_metrics_response, dataset_info_dict, pre_upload_stats)

    if delete_version:
      self.delete_version(dataset_version_id)

  def export(self,
             save_path: str,
             archive_url: str = None,
             local_archive_path: str = None,
             split: str = None,
             num_workers: int = 4) -> None:
    """Exports the Clarifai protobuf dataset to a local archive.

    Args:
        save_path (str): The path to save the archive to.
        archive_url (str): The URL to the Clarifai protobuf archive.
        local_archive_path (str): The path to the local Clarifai protobuf archive.
        split (str): Export dataset inputs in the directory format {split}/{input_type}. Default is all.
        num_workers (int): Number of workers to use for downloading the archive. Default is 4.

    Example:
        >>> from clarifai.client.dataset import Dataset
        >>> Dataset().export(save_path='output.zip', local_archive_path='clarifai-data-protobuf.zip')
    """
    if local_archive_path and not os.path.exists(local_archive_path):
      raise UserError(f"Archive {local_archive_path} does not exist.")
    # Create a session object and set auth header
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
    session.mount('https://', HTTPAdapter(max_retries=retries))
    session.headers.update({'Authorization': self.metadata[0][1]})
    with DatasetExportReader(
        session=session, archive_url=archive_url, local_archive_path=local_archive_path) as reader:
      InputAnnotationDownloader(session, reader, num_workers).download_archive(
          save_path=save_path, split=split)

  def __getattr__(self, name):
    return getattr(self.dataset_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.dataset_info, param)}" for param in init_params
        if hasattr(self.dataset_info, param)
    ]
    return f"Dataset Details: \n{', '.join(attribute_strings)}\n"



import logging
from collections import defaultdict
from typing import Dict, List, Optional, Union

from rich import print as rprint
from rich.logging import RichHandler
from rich.table import Table
from rich.traceback import install
from rich.tree import Tree

install()


def display_workflow_tree(nodes_data: List[Dict]) -> None:
  """Displays a tree of the workflow nodes."""
  # Create a mapping of node_id to the list of node_ids that are connected to it.
  node_adj_mapping = defaultdict(list)
  # Create a mapping of node_id to the node data info.
  nodes_data_dict = {}
  for node in nodes_data:
    nodes_data_dict[node["id"]] = node
    if node.get("node_inputs", "") == "":
      node_adj_mapping["Input"].append(node["id"])
    else:
      for node_input in node["node_inputs"]:
        node_adj_mapping[node_input["node_id"]].append(node["id"])

  # Get all leaf nodes.
  leaf_node_ids = set()
  for node_id in list(nodes_data_dict.keys()):
    if node_adj_mapping.get(node_id, "") == "":
      leaf_node_ids.add(node_id)

  def build_node_tree(node_id="Input"):
    """Recursively builds a rich tree of the workflow nodes."""
    # Set the style of the current node.
    style_str = "green" if node_id in leaf_node_ids else "white"

    # Create a Tree object for the current node.
    if node_id != "Input":
      node_table = table_from_dict(
          [nodes_data_dict[node_id]["model"]],
          column_names=["id", "model_type_id", "app_id", "user_id"],
          title="Node: " + node_id)

      tree = Tree(node_table, style=style_str, guide_style="underline2 white")
    else:
      tree = Tree(f"[green] {node_id}", style=style_str, guide_style="underline2 white")

    # Recursively add the child nodes of the current node to the tree.
    for child in node_adj_mapping.get(node_id, []):
      tree.add(build_node_tree(child))

    # Return the tree.
    return tree

  tree = build_node_tree("Input")
  rprint(tree)


def table_from_dict(data: List[Dict], column_names: List[str], title: str = "") -> Table:
  """Use this function for printing tables from a list of dicts."""
  table = Table(title=title, show_lines=False, show_header=True, header_style="blue")
  for column_name in column_names:
    table.add_column(column_name)
  for row in data:
    req_row = [row.get(column_name, "") for column_name in column_names]
    table.add_row(*req_row)
  return table


def _get_library_name() -> str:
  return __name__.split(".")[0]


def _configure_logger(name: str, logger_level: Union[int, str] = logging.NOTSET) -> None:
  """Configure the logger with the specified name."""

  logger = logging.getLogger(name)
  logger.setLevel(logger_level)

  # Remove existing handlers
  for handler in logger.handlers[:]:
    logger.removeHandler(handler)

  # Add the new rich handler and formatter
  handler = RichHandler(rich_tracebacks=True, log_time_format="%Y-%m-%d %H:%M:%S")
  formatter = logging.Formatter('%(name)s:  %(message)s')
  handler.setFormatter(formatter)
  logger.addHandler(handler)


def get_logger(logger_level: Union[int, str] = logging.NOTSET,
               name: Optional[str] = None) -> logging.Logger:
  """Return a logger with the specified name."""

  if name is None:
    name = _get_library_name()

  _configure_logger(name, logger_level)
  return logging.getLogger(name)


def add_file_handler(logger: logging.Logger, file_path: str, log_level: str = 'WARNING') -> None:
  """Add a file handler to the logger."""
  file_handler = logging.FileHandler(file_path)
  file_handler.setLevel(log_level)
  logger.addHandler(file_handler)



import os

from .constants import CLARIFAI_PAT_PATH


def _persist_pat(pat: str):
  """ Write down pat to CLARIFAI_PAT_PATH """
  with open(CLARIFAI_PAT_PATH, "w") as f:
    f.write(pat)


def _read_pat():
  if not os.path.exists(CLARIFAI_PAT_PATH):
    return None
  with open(CLARIFAI_PAT_PATH, "r") as f:
    return f.read().replace("\n", "").replace("\r", "").strip()


def login(pat=None):
  """ if pat provided, set pat to CLARIFAI_PAT otherwise read pat from file"""
  os.environ["CLARIFAI_PAT"] = pat or _read_pat()



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Commandline interface for model upload utils."""
import argparse
import os
import subprocess
from typing import Union

from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.models.api import Models
from clarifai.models.model_serving.model_config import (MODEL_TYPES, get_model_config,
                                                        load_user_config)
from clarifai.models.model_serving.model_config.inference_parameter import InferParamManager

from ..utils import login
from .base import BaseClarifaiCli


class UploadCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: argparse._SubParsersAction):
    creator_parser = parser.add_parser("upload", help="Upload component to Clarifai platform")
    sub_creator_parser = creator_parser.add_subparsers()

    UploadModelSubCli.register(sub_creator_parser)

    creator_parser.set_defaults(func=UploadCli)


class UploadModelSubCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: argparse._SubParsersAction):
    upload_parser = parser.add_parser("model", help="Upload Clarifai model")
    upload_parser.add_argument(
        "path",
        type=str,
        nargs='?',
        help=
        "Path to working dir to get clarifai_config.yaml or path to yaml. Default is current directory",
        default=".")
    upload_parser.add_argument(
        "--url", type=str, required=True, help="Direct download url of zip file")
    upload_parser.add_argument("--id", type=str, required=False, help="Model ID")
    upload_parser.add_argument(
        "--user-app",
        type=str,
        required=False,
        help="User ID and App ID separated by '/', e.g., <user_id>/<app_id>")
    upload_parser.add_argument(
        "--no-test",
        action="store_true",
        help="Trigger this flag to skip testing before uploading")

    upload_parser.add_argument(
        "--update-version",
        action="store_true",
        required=False,
        help="Update exist model with new version")

    upload_parser.set_defaults(func=UploadModelSubCli)

  def __init__(self, args: argparse.Namespace) -> None:
    self.no_test = args.no_test

    working_dir_or_config = args.path
    # if input a config file, then not running test
    if working_dir_or_config.endswith(".yaml"):
      config_yaml_path = working_dir_or_config
      self.test_path = None
      self.no_test = True
    # if it is a directory - working dir then it must contain config and test
    else:
      config_yaml_path = os.path.join(working_dir_or_config, "clarifai_config.yaml")
      self.test_path = os.path.join(working_dir_or_config, "test.py")

    assert os.path.exists(config_yaml_path), FileNotFoundError(
        f"`{config_yaml_path}` does not exist")
    self.config = load_user_config(cfg_path=config_yaml_path)

    self.user_id, self.app_id = "", ""
    user_app = args.user_app
    self.url: str = args.url
    self.update_version = args.update_version
    assert self.url.startswith("http") or self.url.startswith(
        "s3"), f"Invalid url supported http or s3 url. Got {self.url}"

    clarifai_cfg = self.config.clarifai_model
    self.url: str = args.url
    self.id = args.id or clarifai_cfg.clarifai_model_id
    self.type = clarifai_cfg.type
    self.desc = clarifai_cfg.description
    self.infer_param = clarifai_cfg.inference_parameters
    user_app = user_app or clarifai_cfg.clarifai_user_app_id

    if user_app:
      user_app = user_app.split('/')
      assert len(
          user_app
      ) == 2, f"id must be combination of user_id and app_id separated by `/`, e.g. <user_id>/<app_id>. Got {args.id}"
      self.user_id, self.app_id = user_app

    if self.user_id:
      os.environ["CLARIFAI_USER_ID"] = self.user_id
    if self.app_id:
      os.environ["CLARIFAI_APP_ID"] = self.app_id

    _user_id = os.environ.get("CLARIFAI_USER_ID", None)
    _app_id = os.environ.get("CLARIFAI_APP_ID", None)
    assert _user_id or _app_id, f"Missing user-id or app-id, got user-id {_user_id} and app-id {_app_id}"
    login()

  def run(self):

    # Run test before uploading
    if not self.no_test:
      assert os.path.exists(self.test_path), FileNotFoundError(f"Not found {self.test_path}")
      result = subprocess.run(f"pytest -s --log-level=INFO {self.test_path}")
      assert result.returncode == 0, "Test has failed. Please make sure no error exists in your code."

    deploy(
        model_url=self.url,
        model_id=self.id,
        desc=self.desc,
        model_type=self.type,
        update_version=self.update_version,
        inference_parameters=self.infer_param)


def deploy(model_url,
           model_id: str = None,
           model_type: str = None,
           desc: str = "",
           update_version: bool = False,
           inference_parameters: Union[dict, str] = None):
  # init Auth from env vars
  auth = ClarifaiAuthHelper.from_env()
  # init api
  model_api = Models(auth)
  # key map
  assert model_type in MODEL_TYPES, f"model_type should be one of {MODEL_TYPES}"
  clarifai_key_map = get_model_config(model_type=model_type).clarifai_model.field_maps
  # inference parameters
  if isinstance(inference_parameters, str) and os.path.isfile(inference_parameters):
    inference_parameters = InferParamManager(json_path=inference_parameters).get_list_params()
  # if updating new version of existing model
  if update_version:
    resp = model_api.post_model_version(
        model_id=model_id,
        model_zip_url=model_url,
        input=clarifai_key_map.input_fields_map,
        outputs=clarifai_key_map.output_fields_map,
        param_specs=inference_parameters)
  # creating new model
  else:
    # post model
    resp = model_api.upload_model(
        model_id=model_id,
        model_zip_url=model_url,
        model_type=model_type,
        input=clarifai_key_map.input_fields_map,
        outputs=clarifai_key_map.output_fields_map,
        description=desc,
        param_specs=inference_parameters)
  # response
  if resp["status"]["code"] != "SUCCESS":
    raise Exception("Post models failed, details: {}, {}".format(resp["status"]["description"],
                                                                 resp["status"]["details"]))
  else:
    print("Success!")
    print(f'Model version: {resp["model"]["model_version"]["id"]}')



import os
import time
from typing import Any, Dict, Generator, List, Union

import requests
import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.resources_pb2 import Input
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.model import MAX_MODEL_PREDICT_INPUTS, TRAINABLE_MODEL_TYPES
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import get_logger
from clarifai.utils.misc import BackoffIterator
from clarifai.utils.model_train import (find_and_replace_key, params_parser,
                                        response_to_model_params, response_to_param_info,
                                        response_to_templates)


class Model(Lister, BaseClient):
  """Model is a class that provides access to Clarifai API endpoints related to Model information."""

  def __init__(self,
               url: str = None,
               model_id: str = None,
               model_version: Dict = {'id': ""},
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               **kwargs):
    """Initializes a Model object.

    Args:
        url (str): The URL to initialize the model object.
        model_id (str): The Model ID to interact with.
        model_version (dict): The Model Version to interact with.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        **kwargs: Additional keyword arguments to be passed to the Model.
    """
    if url and model_id:
      raise UserError("You can only specify one of url or model_id.")
    if not url and not model_id:
      raise UserError("You must specify one of url or model_id.")
    if url:
      user_id, app_id, _, model_id, model_version_id = ClarifaiUrlHelper.split_clarifai_url(url)
      model_version = {'id': model_version_id}
      kwargs = {'user_id': user_id, 'app_id': app_id}
    self.kwargs = {**kwargs, 'id': model_id, 'model_version': model_version,}
    self.model_info = resources_pb2.Model(**self.kwargs)
    self.logger = get_logger(logger_level="INFO", name=__name__)
    self.training_params = {}
    BaseClient.__init__(
        self, user_id=self.user_id, app_id=self.app_id, base=base_url, pat=pat, token=token)
    Lister.__init__(self)

  def list_training_templates(self) -> List[str]:
    """Lists all the training templates for the model type.

    Returns:
        templates (List): List of training templates for the model type.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> print(model.list_training_templates())
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    templates = response_to_templates(
        response=response, model_type_id=self.model_info.model_type_id)

    return templates

  def get_params(self, template: str = None, save_to: str = 'params.yaml') -> Dict[str, Any]:
    """Returns the model params for the model type and yaml file.

    Args:
        template (str): The template to use for the model type.
        yaml_file (str): The yaml file to save the model params.

    Returns:
        params (Dict): Dictionary of model params for the model type.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if template is None and self.model_info.model_type_id not in [
        "clusterer", "embedding-classifier"
    ]:
      raise UserError(
          f"Template should be provided for {self.model_info.model_type_id} model type")
    if template is not None and self.model_info.model_type_id in [
        "clusterer", "embedding-classifier"
    ]:
      raise UserError(
          f"Template should not be provided for {self.model_info.model_type_id} model type")

    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    params = response_to_model_params(
        response=response, model_type_id=self.model_info.model_type_id, template=template)
    #yaml file
    assert save_to.endswith('.yaml'), "File extension should be .yaml"
    with open(save_to, 'w') as f:
      yaml.dump(params, f, default_flow_style=False, sort_keys=False)
    #updating the global model params
    self.training_params.update(params)

    return params

  def update_params(self, **kwargs) -> None:
    """Updates the model params for the model.

    Args:
        **kwargs: model params to update.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.update_params(batch_size = 8, dataset_version = 'dataset_version_id')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if len(self.training_params) == 0:
      raise UserError(
          f"Run 'model.get_params' to get the params for the {self.model_info.model_type_id} model type"
      )
    #getting all the keys in nested dictionary
    all_keys = [key for key in self.training_params.keys()] + [
        key for key in self.training_params.values() if isinstance(key, dict) for key in key
    ]
    #checking if the given params are valid
    if not set(kwargs.keys()).issubset(all_keys):
      raise UserError("Invalid params")
    #updating the global model params
    for key, value in kwargs.items():
      find_and_replace_key(self.training_params, key, value)

  def get_param_info(self, param: str) -> Dict[str, Any]:
    """Returns the param info for the param.

    Args:
        param (str): The param to get the info for.

    Returns:
        param_info (Dict): Dictionary of model param info for the param.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.get_param_info('param')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if len(self.training_params) == 0:
      raise UserError(
          f"Run 'model.get_params' to get the params for the {self.model_info.model_type_id} model type"
      )

    all_keys = [key for key in self.training_params.keys()] + [
        key for key in self.training_params.values() if isinstance(key, dict) for key in key
    ]
    if param not in all_keys:
      raise UserError(f"Invalid param: '{param}' for model type '{self.model_info.model_type_id}'")
    template = self.training_params['train_params']['template'] if 'template' in all_keys else None

    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    param_info = response_to_param_info(
        response=response,
        model_type_id=self.model_info.model_type_id,
        param=param,
        template=template)

    return param_info

  def train(self, yaml_file: str = None) -> str:
    """Trains the model based on the given yaml file or model params.

    Args:
        yaml_file (str): The yaml file for the model params.

    Returns:
        model_version_id (str): The model version ID for the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.train('model_params.yaml')
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if not yaml_file and len(self.training_params) == 0:
      raise UserError("Provide yaml file or run 'model.get_params()'")

    if yaml_file:
      with open(yaml_file, 'r') as file:
        params_dict = yaml.safe_load(file)
    else:
      params_dict = self.training_params
    #getting all the concepts for the model type
    if self.model_info.model_type_id not in ["clusterer", "text-to-text"]:
      concepts = self._list_concepts()
    train_dict = params_parser(params_dict, concepts)
    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[resources_pb2.ModelVersion(**train_dict)])
    response = self._grpc_request(self.STUB.PostModelVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Training Started\n%s", response.status)

    return response.model.model_version.id

  def training_status(self, version_id: str, training_logs: bool = False) -> Dict[str, str]:
    """Get the training status for the model version. Also stores training logs

    Args:
        version_id (str): The version ID to get the training status for.
        training_logs (bool): Whether to save the training logs in a file.

    Returns:
        training_status (Dict): Dictionary of training status for the model version.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.training_status(version_id='version_id',training_logs=True)
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")

    request = service_pb2.GetModelVersionRequest(
        user_app_id=self.user_app_id, model_id=self.id, version_id=version_id)
    response = self._grpc_request(self.STUB.GetModelVersion, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    if training_logs:
      try:
        if response.model_version.train_log:
          log_response = requests.get(response.model_version.train_log)
          log_response.raise_for_status()  # Check for any HTTP errors
          with open(version_id + '.log', 'wb') as file:
            for chunk in log_response.iter_content(chunk_size=4096):  # 4KB
              file.write(chunk)
          self.logger.info(f"\nTraining logs are saving in '{version_id+'.log'}' file")

      except requests.exceptions.RequestException as e:
        raise Exception(f"An error occurred while getting training logs: {e}")

    return response.model_version.status

  def delete_version(self, version_id: str) -> None:
    """Deletes a model version for the Model.

    Args:
        version_id (str): The version ID to delete.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.delete_version(version_id='version_id')
    """
    request = service_pb2.DeleteModelVersionRequest(
        user_app_id=self.user_app_id, model_id=self.id, version_id=version_id)

    response = self._grpc_request(self.STUB.DeleteModelVersion, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Version Deleted\n%s", response.status)

  def create_version(self, **kwargs) -> 'Model':
    """Creates a model version for the Model.

    Args:
        **kwargs: Additional keyword arguments to be passed to Model Version.
          - description (str): The description of the model version.
          - concepts (list[Concept]): The concepts to associate with the model version.
          - output_info (resources_pb2.OutputInfo(): The output info to associate with the model version.

    Returns:
        Model: A Model object for the specified model ID.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url")
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_version = model.create_version(description='model_version_description')
    """
    if self.model_info.model_type_id in TRAINABLE_MODEL_TYPES:
      raise UserError(
          f"{self.model_info.model_type_id} is a trainable model type. Use 'model.train()' to train the model"
      )

    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[resources_pb2.ModelVersion(**kwargs)])

    response = self._grpc_request(self.STUB.PostModelVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Version created\n%s", response.status)

    kwargs.update({'app_id': self.app_id, 'user_id': self.user_id})
    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response['model'], 'model')

    return Model(base_url=self.base, pat=self.pat, token=self.token, **kwargs)

  def list_versions(self, page_no: int = None,
                    per_page: int = None) -> Generator['Model', None, None]:
    """Lists all the versions for the model.

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Model: Model objects for the versions of the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> all_model_versions = list(model.list_versions())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        model_id=self.id,
    )
    all_model_versions_info = self.list_pages_generator(
        self.STUB.ListModelVersions,
        service_pb2.ListModelVersionsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for model_version_info in all_model_versions_info:
      model_version_info['id'] = model_version_info['model_version_id']
      del model_version_info['model_version_id']
      try:
        del model_version_info['train_info']['dataset']['version']['metrics']
      except KeyError:
        pass
      yield Model.from_auth_helper(
          model_id=self.id, **dict(self.kwargs, model_version=model_version_info))

  def predict(self, inputs: List[Input], inference_params: Dict = {}, output_config: Dict = {}):
    """Predicts the model based on the given inputs.

    Args:
        inputs (list[Input]): The inputs to predict, must be less than 128.
    """
    if not isinstance(inputs, list):
      raise UserError('Invalid inputs, inputs must be a list of Input objects.')
    if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
      raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}."
                     )  # TODO Use Chunker for inputs len > 128

    self._override_model_version(inference_params, output_config)
    request = service_pb2.PostModelOutputsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_version.id,
        inputs=inputs,
        model=self.model_info)

    start_time = time.time()
    backoff_iterator = BackoffIterator()
    while True:
      response = self._grpc_request(self.STUB.PostModelOutputs, request)

      if response.status.code == status_code_pb2.MODEL_DEPLOYING and \
        time.time() - start_time < 60 * 10: # 10 minutes
        self.logger.info(f"{self.id} model is still deploying, please wait...")
        time.sleep(next(backoff_iterator))
        continue

      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model Predict failed with response {response.status!r}")
      else:
        break

    return response

  def predict_by_filepath(self,
                          filepath: str,
                          input_type: str,
                          inference_params: Dict = {},
                          output_config: Dict = {}):
    """Predicts the model based on the given filepath.

    Args:
        filepath (str): The filepath to predict.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_prediction = model.predict_by_filepath('/path/to/image.jpg', 'image')
        >>> model_prediction = model.predict_by_filepath('/path/to/text.txt', 'text')
    """
    if not os.path.isfile(filepath):
      raise UserError('Invalid filepath.')

    with open(filepath, "rb") as f:
      file_bytes = f.read()

    return self.predict_by_bytes(file_bytes, input_type, inference_params, output_config)

  def predict_by_bytes(self,
                       input_bytes: bytes,
                       input_type: str,
                       inference_params: Dict = {},
                       output_config: Dict = {}):
    """Predicts the model based on the given bytes.

    Args:
        input_bytes (bytes): File Bytes to predict on.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("https://clarifai.com/openai/chat-completion/models/GPT-4")
        >>> model_prediction = model.predict_by_bytes(b'Write a tweet on future of AI',
                                                      input_type='text',
                                                      inference_params=dict(temperature=str(0.7), max_tokens=30)))
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError(
          f"Got input type {input_type} but expected one of image, text, video, audio.")
    if not isinstance(input_bytes, bytes):
      raise UserError('Invalid bytes.')

    if input_type == "image":
      input_proto = Inputs.get_input_from_bytes("", image_bytes=input_bytes)
    elif input_type == "text":
      input_proto = Inputs.get_input_from_bytes("", text_bytes=input_bytes)
    elif input_type == "video":
      input_proto = Inputs.get_input_from_bytes("", video_bytes=input_bytes)
    elif input_type == "audio":
      input_proto = Inputs.get_input_from_bytes("", audio_bytes=input_bytes)

    return self.predict(
        inputs=[input_proto], inference_params=inference_params, output_config=output_config)

  def predict_by_url(self,
                     url: str,
                     input_type: str,
                     inference_params: Dict = {},
                     output_config: Dict = {}):
    """Predicts the model based on the given URL.

    Args:
        url (str): The URL to predict.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_prediction = model.predict_by_url('url', 'image')
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError(
          f"Got input type {input_type} but expected one of image, text, video, audio.")

    if input_type == "image":
      input_proto = Inputs.get_input_from_url("", image_url=url)
    elif input_type == "text":
      input_proto = Inputs.get_input_from_url("", text_url=url)
    elif input_type == "video":
      input_proto = Inputs.get_input_from_url("", video_url=url)
    elif input_type == "audio":
      input_proto = Inputs.get_input_from_url("", audio_url=url)

    return self.predict(
        inputs=[input_proto], inference_params=inference_params, output_config=output_config)

  def _override_model_version(self, inference_params: Dict = {}, output_config: Dict = {}) -> None:
    """Overrides the model version.

    Args:
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.
          sample_ms (int): The number of milliseconds to sample.
    """
    params = Struct()
    if inference_params is not None:
      params.update(inference_params)

    self.model_info.model_version.output_info.CopyFrom(
        resources_pb2.OutputInfo(
            output_config=resources_pb2.OutputConfig(**output_config), params=params))

  def _list_concepts(self) -> List[str]:
    """Lists all the concepts for the model type.

    Returns:
        concepts (List): List of concepts for the model type.
    """
    request_data = dict(user_app_id=self.user_app_id)
    all_concepts_infos = self.list_pages_generator(self.STUB.ListConcepts,
                                                   service_pb2.ListConceptsRequest, request_data)
    return [concept_info['concept_id'] for concept_info in all_concepts_infos]

  def load_info(self) -> None:
    """Loads the model info."""
    request = service_pb2.GetModelRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_info.model_version.id)
    response = self._grpc_request(self.STUB.GetModel, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    self.kwargs = self.process_response_keys(dict_response['model'])
    self.model_info = resources_pb2.Model(**self.kwargs)

  def __getattr__(self, name):
    return getattr(self.model_info, name)

  def __str__(self):
    if len(self.kwargs) < 10:
      self.load_info()

    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.model_info, param)}" for param in init_params
        if hasattr(self.model_info, param)
    ]
    return f"Model Details: \n{', '.join(attribute_strings)}\n"

  def list_evaluations(self) -> resources_pb2.EvalMetrics:
    """List all eval_metrics of current model version

    Raises:
        Exception: Failed to call API

    Returns:
        resources_pb2.EvalMetrics
    """
    assert self.model_info.model_version.id, "Model version is empty. Please provide `model_version` as arguments or with a URL as the format '{user_id}/{app_id}/models/{your_model_id}/model_version_id/{your_version_model_id}' when initializing."
    request = service_pb2.ListModelVersionEvaluationsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_version_id=self.model_info.model_version.id)
    response = self._grpc_request(self.STUB.ListModelVersionEvaluations, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    return response.eval_metrics

  def evaluate(self,
               dataset_id: str,
               dataset_app_id: str = None,
               dataset_user_id: str = None,
               eval_id: str = None,
               extended_metrics: dict = None,
               eval_info: dict = None) -> resources_pb2.EvalMetrics:
    """ Run evaluation

    Args:
      dataset_id (str): Dataset Id.
      dataset_app_id (str): App ID for cross app evaluation, leave it as None to use Model App ID. Default is None.
      dataset_user_id (str): User ID for cross app evaluation, leave it as None to use Model User ID. Default is None.
      eval_id (str): Specific ID for the evaluation. You must specify this parameter to either overwrite the result with the dataset ID or format your evaluation in an informative manner. If you don't, it will use random ID from system. Default is None.
      extended_metrics (dict): user custom metrics result. Default is None.
      eval_info (dict): custom eval info. Default is empty dict.

    Return
      eval_metrics

    """
    assert self.model_info.model_version.id, "Model version is empty. Please provide `model_version` as arguments or with a URL as the format '{user_id}/{app_id}/models/{your_model_id}/model_version_id/{your_version_model_id}' when initializing."
    metrics = None
    if isinstance(extended_metrics, dict):
      metrics = Struct()
      metrics.update(extended_metrics)
      metrics = resources_pb2.ExtendedMetrics(user_metrics=metrics)

    eval_info_params = None
    if isinstance(eval_info, dict):
      eval_info_params = Struct()
      eval_info_params.update(eval_info)
      eval_info_params = resources_pb2.EvalInfo(params=eval_info_params)

    eval_metric = resources_pb2.EvalMetrics(
        id=eval_id,
        model=resources_pb2.Model(
            id=self.id,
            app_id=self.auth_helper.app_id,
            user_id=self.auth_helper.user_id,
            model_version=resources_pb2.ModelVersion(id=self.model_info.model_version.id),
        ),
        extended_metrics=metrics,
        ground_truth_dataset=resources_pb2.Dataset(
            id=dataset_id,
            app_id=dataset_app_id or self.auth_helper.app_id,
            user_id=dataset_user_id or self.auth_helper.user_id,
        ),
        eval_info=eval_info_params,
    )
    request = service_pb2.PostEvaluationsRequest(
        user_app_id=self.user_app_id,
        eval_metrics=[eval_metric],
    )
    response = self._grpc_request(self.STUB.PostEvaluations, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info(
        "\nModel evaluation in progress. Kindly allow a few minutes for completion. Processing time may vary based on the model and dataset sizes."
    )

    return response.eval_metrics

  def get_eval_by_id(
      self,
      eval_id: str,
      label_counts=False,
      test_set=False,
      binary_metrics=False,
      confusion_matrix=False,
      metrics_by_class=False,
      metrics_by_area=False,
  ) -> resources_pb2.EvalMetrics:
    """Get detail eval_metrics by eval_id with extra metric fields

    Args:
        eval_id (str): eval id
        label_counts (bool, optional): Set True to get label counts. Defaults to False.
        test_set (bool, optional): Set True to get test set. Defaults to False.
        binary_metrics (bool, optional): Set True to get binary metric. Defaults to False.
        confusion_matrix (bool, optional): Set True to get confusion matrix. Defaults to False.
        metrics_by_class (bool, optional): Set True to get metrics by class. Defaults to False.
        metrics_by_area (bool, optional): Set True to get metrics by area. Defaults to False.

    Raises:
        Exception: Failed to call API

    Returns:
        resources_pb2.EvalMetrics: eval_metrics
    """
    request = service_pb2.GetEvaluationRequest(
        user_app_id=self.user_app_id,
        evaluation_id=eval_id,
        fields=resources_pb2.FieldsValue(
            label_counts=label_counts,
            test_set=test_set,
            binary_metrics=binary_metrics,
            confusion_matrix=confusion_matrix,
            metrics_by_class=metrics_by_class,
            metrics_by_area=metrics_by_area,
        ))
    response = self._grpc_request(self.STUB.GetEvaluation, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    return response.eval_metrics

  def get_latest_eval(self,
                      label_counts=False,
                      test_set=False,
                      binary_metrics=False,
                      confusion_matrix=False,
                      metrics_by_class=False,
                      metrics_by_area=False) -> Union[resources_pb2.EvalMetrics, None]:
    """
    Run `get_eval_by_id` method with latest `eval_id`

    Args:
      label_counts (bool, optional): Set True to get label counts. Defaults to False.
      test_set (bool, optional): Set True to get test set. Defaults to False.
      binary_metrics (bool, optional): Set True to get binary metric. Defaults to False.
      confusion_matrix (bool, optional): Set True to get confusion matrix. Defaults to False.
      metrics_by_class (bool, optional): Set True to get metrics by class. Defaults to False.
      metrics_by_area (bool, optional): Set True to get metrics by area. Defaults to False.

    Returns:
      eval_metric if model is evaluated otherwise None.

    """

    _latest = self.list_evaluations()[0]
    result = None
    if _latest.status.code == status_code_pb2.MODEL_EVALUATED:
      result = self.get_eval_by_id(
          eval_id=_latest.id,
          label_counts=label_counts,
          test_set=test_set,
          binary_metrics=binary_metrics,
          confusion_matrix=confusion_matrix,
          metrics_by_class=metrics_by_class,
          metrics_by_area=metrics_by_area)

    return result



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface to Clarifai Runners API."""

import os
import threading
import traceback
from concurrent.futures import ThreadPoolExecutor

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf import json_format

from clarifai.client.base import BaseClient
from clarifai.errors import UserError
from clarifai.utils.logging import get_logger


class Runner(BaseClient):
  """Base class for remote inference runners. This should be subclassed with the run_input method
  implemented to process each input in the request.

  Then on the subclass call start() to start the run loop.
  """

  def __init__(self,
               runner_id: str,
               user_id: str = None,
               check_runner_exists: bool = True,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               num_parallel_polls: int = 4,
               **kwargs) -> None:
    """
    Args:
      runner_id (str): the id of the runner to use. Create the runner in the Clarifai API first
      user_id (str): Clarifai User ID
      base_url (str): Base API url. Default "https://api.clarifai.com"
      pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
      token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
      num_parallel_polls (int): the max number of threads for parallel run loops to be fetching work from
    """
    user_id = user_id or os.environ.get("CLARIFAI_USER_ID", "")

    if not user_id:
      raise UserError(
          "Set CLARIFAI_USER_ID as environment variables or pass user_id as input arguments")

    self.runner_id = runner_id
    self.logger = get_logger("INFO", __name__)
    self.kwargs = {**kwargs, 'id': runner_id, 'user_id': user_id}
    self.runner_info = resources_pb2.Runner(**self.kwargs)
    self.num_parallel_polls = min(10, num_parallel_polls)
    BaseClient.__init__(self, user_id=self.user_id, app_id="", base=base_url, pat=pat, token=token)

    # Check that the runner exists.
    if check_runner_exists:
      request = service_pb2.GetRunnerRequest(user_app_id=self.user_app_id, runner_id=runner_id)
      response = self._grpc_request(self.STUB.GetRunner, request)
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(
            f"Error getting runner, are you use this is a valid runner id {runner_id} for user_id {user_id}. Error: {response.status.description}, Details: {response.status.details}"
        )

  def start(self):
    """
    Start the run loop. This will ask the Clarifai API for work, and when it gets work, it will run
    the model on the inputs and post the results back to the Clarifai API. It will then ask for more
    work again.
    """
    self._long_poll_loop()

  def _run(self, request: service_pb2.PostModelOutputsRequest) -> service_pb2.MultiOutputResponse:
    """
    Run the model on the given request. You shouldn't need to override this method, see run_input
    for the implementation to process each input in the request.

    Args:
      request: service_pb2.PostModelOutputsRequest - the request to run the model on

    Returns:
      service_pb2.MultiOutputResponse - the response from the model's run_input implementation.
    """
    outputs = []

    # TODO: Could cache the model and this conversion if the hash is the same.
    model = request.model
    output_info = None
    if request.model.model_version.id != "":
      output_info = json_format.MessageToDict(
          model.model_version.output_info, preserving_proto_field_name=True)

    # TODO: parallelize this over inputs in a single request.
    for inp in request.inputs:
      try:
        output = self.run_input(inp, output_info)
      except Exception as e:
        self.logger.exception("run_input failed:")
        traceback.print_exc()
        output = resources_pb2.Output(status=status_pb2.Status(
            code=status_code_pb2.MODEL_PREDICTION_FAILED,
            description="Failed",
            details="",
            stack_trace=traceback.format_exc().split('\n'),
            internal_details=str(e),
        ))
        self.logger.exception(output)
      outputs.append(output)

    successes = [o.status.code == status_code_pb2.SUCCESS for o in outputs]
    if all(successes):
      status = status_pb2.Status(
          code=status_code_pb2.SUCCESS,
          description="Success",
      )
    elif any(successes):
      status = status_pb2.Status(
          code=status_code_pb2.MIXED_STATUS,
          description="Mixed Status",
      )
    else:
      status = status_pb2.Status(
          code=status_code_pb2.FAILURE,
          description="Failed",
      )

    return service_pb2.MultiOutputResponse(
        status=status,
        outputs=outputs,
    )

  def run_input(self, input: resources_pb2.Input,
                output_info: resources_pb2.OutputInfo) -> resources_pb2.Output:
    """
    Run the model on the given input in the request. This is the method you should override to
    process each input in the request.

    Args:
      input: resources_pb2.Input - the input to run the model on
      output_info: resources_pb2.OutputInfo - the output info for the model which includes
    output_info.params that the model can pass in on very prediction request. These can be provided
    during PostModelVersions as default for every request or can be overridden on a per request by
    passing in output_info in the PostModelOutputs request as the model.model_version.output_info.params
    field.

    Returns:
      resources_pb2.Output - the response from the model's run_input implementation.
    """
    raise NotImplementedError("run_input() not implemented")

  def _long_poll_loop(self):
    """
    This method will long poll for work, and when it gets work, it will run the model on the inputs
    and post the results back to the Clarifai API. It will then long poll again for more work.
    """

    def _single_thread_loop():
      c = 0
      while True:
        # Long poll waiting for work.
        try:
          self.logger.info(f"Loop iteration: {c} for thread {threading.get_ident()}")
          work_response = self._grpc_request(self.STUB.ListRunnerItems,
                                             service_pb2.ListRunnerItemsRequest(
                                                 user_app_id=self.user_app_id,
                                                 runner_id=self.runner_id))
          if work_response.status.code == status_code_pb2.RUNNER_NEEDS_RETRY:
            c += 1
            continue  # immediate restart the long poll
          if work_response.status.code != status_code_pb2.SUCCESS:
            raise Exception("Error getting work: {}".format(work_response.status.description))
          if len(work_response.items) == 0:
            self.logger.info("No work to do. Waiting...")
            continue

          # We have work to do. Run the model on each item fetched (typically 1 item at a time right
          # now is returned by ListRunnerItems).
          for item in work_response.items:
            if not item.HasField('post_model_outputs_request'):
              raise Exception("Unexpected work item type: {}".format(item))
            self.logger.info(
                f"Working on item: {item.id} with inputs {len(item.post_model_outputs_request.inputs)}"
            )
            # run this batch of data through the model.
            result = self._run(item.post_model_outputs_request)

            result_response = self._grpc_request(
                self.STUB.PostRunnerItemOutputs,
                service_pb2.PostRunnerItemOutputsRequest(
                    user_app_id=self.user_app_id,
                    item_id=item.id,
                    runner_id=self.runner_id,
                    runner_item_outputs=[
                        service_pb2.RunnerItemOutput(multi_output_response=result)
                    ]))
            if result_response.status.code != status_code_pb2.SUCCESS:
              raise Exception(
                  json_format.MessageToJson(result_response, preserving_proto_field_name=True))
              # raise Exception("Error posting result: {}".format(result_response.status.description))
        except Exception as e:  # keep the loop running
          self.logger.exception(e)
          continue

    self.logger.info(f"Starting run loop for {self.num_parallel_polls} threads...")
    if self.num_parallel_polls == 1:
      _single_thread_loop()
    else:
      pool = ThreadPoolExecutor(max_workers=self.num_parallel_polls)
      futures = [pool.submit(_single_thread_loop) for _ in range(self.num_parallel_polls)]
      for f in futures:
        self.logger.info(f.result())
      pool.shutdown()

  def __getattr__(self, name):
    return getattr(self.runner_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.runner_info, param)}" for param in init_params
        if hasattr(self.runner_info, param)
    ]
    return f"Runner Details: \n{', '.join(attribute_strings)}\n"



import time

import requests
from clarifai_grpc.grpc.api import resources_pb2
from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline

from clarifai.client.runner import Runner

# This example requires to run the following before running this example:
# pip install transformers

# https://huggingface.co/TheBloke/Llama-2-70B-chat-GPTQ
model_name_or_path = "TheBloke/Llama-2-7B-chat-GPTQ"
model_basename = "model"

use_triton = False


class Llama2Runner(Runner):
  """A custom runner that runs the LLama2 LLM.
  """

  def __init__(self, *args, **kwargs):
    super(Llama2Runner, self).__init__(*args, **kwargs)
    self.logger.info("Starting to load the model...")
    st = time.time()
    self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, use_fast=True)
    self.model = AutoModelForCausalLM.from_pretrained(model_name_or_path, device_map='auto')

    self.logger.info("Loading model complete in (%f seconds), ready to loop for requests." %
                     (time.time() - st))

  def run_input(self, input: resources_pb2.Input,
                output_info: resources_pb2.OutputInfo) -> resources_pb2.Output:
    """This is the method that will be called when the runner is run. It takes in an input and
    returns an output.
    """

    output = resources_pb2.Output()
    data = input.data
    if data.text.raw != "":
      input_text = data.text.raw
    elif data.text.url != "":
      input_text = str(requests.get(data.text.url).text)
    else:
      raise Exception("Need to include data.text.raw or data.text.url in your inputs.")

    if "params" in output_info:
      params_dict = output_info["params"]
      self.logger.info("params_dict: %s", params_dict)

    time.time()
    max_tokens = 1024
    # # Method 1
    # input_ids = self.tokenizer(input_text, return_tensors='pt').input_ids.cuda()
    # out = self.model.generate(inputs=input_ids, temperature=0.7, max_new_tokens=max_tokens)
    # out_text = self.tokenizer.decode(out[0], skip_special_tokens=True)
    # output.data.text.raw = out_text.replace(input_text, '')

    # # Method 2
    pipe = pipeline(
        "text-generation",
        model=self.model,
        tokenizer=self.tokenizer,
        max_new_tokens=max_tokens,
        temperature=0.7,
        top_p=0.95,
        repetition_penalty=1.15,
        return_full_text=False)
    a = pipe(input_text)
    output.data.text.raw = a[0]['generated_text']
    return output


if __name__ == '__main__':
  # Make sure you set these env vars before running the example.
  # CLARIFAI_PAT
  # CLARIFAI_USER_ID

  # You need to first create a runner in the Clarifai API and then use the ID here.
  Llama2Runner(runner_id="sdk-llama2-runner").start()



from clarifai_grpc.grpc.api import resources_pb2

from clarifai.client.runner import Runner


class MyRunner(Runner):
  """A custom runner that adds "Hello World" to the end of the text and replaces the domain of the
  image URL as an example.
  """

  def run_input(self, input: resources_pb2.Input,
                output_info: resources_pb2.OutputInfo) -> resources_pb2.Output:
    """This is the method that will be called when the runner is run. It takes in an input and
    returns an output.
    """

    output = resources_pb2.Output()

    data = input.data

    # Optional use of output_info
    params_dict = {}
    if "params" in output_info:
      params_dict = output_info["params"]

    if data.text.raw != "":
      output.data.text.raw = data.text.raw + "Hello World" + params_dict.get("hello", "")
    if data.image.url != "":
      output.data.text.raw = data.image.url.replace("samples.clarifai.com",
                                                    "newdomain.com" + params_dict.get("domain",))
    return output


if __name__ == '__main__':
  # Make sure you set these env vars before running the example.
  # CLARIFAI_PAT
  # CLARIFAI_USER_ID

  # You need to first create a runner in the Clarifai API and then use the ID here.
  MyRunner(runner_id="sdk-test-runner").start()



import argparse

from ..constants import CLARIFAI_PAT_PATH
from ..utils import _persist_pat
from .base import BaseClarifaiCli


class LoginCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: argparse._SubParsersAction):
    upload_parser = parser.add_parser("login", help="Login to Clarifai and save PAT locally")
    upload_parser.set_defaults(func=LoginCli)

  def __init__(self, args: argparse.Namespace) -> None:
    pass

  def _parse_config(self):
    # do something with self.config_path
    raise NotImplementedError()

  def run(self):
    msg = "Get your PAT from https://clarifai.com/settings/security and pass it here: "
    _pat = input(msg)
    _persist_pat(_pat)
    print(f"Your PAT is saved at {CLARIFAI_PAT_PATH}")



from abc import ABC, abstractmethod
from argparse import _SubParsersAction


class BaseClarifaiCli(ABC):

  @staticmethod
  @abstractmethod
  def register(parser: _SubParsersAction):
    raise NotImplementedError()

  @abstractmethod
  def run(self):
    raise NotImplementedError()
